#ifndef SPLIT_H
#define SPLIT_H

#include<vcg/complex/append.h>
namespace vcg
{

class Split
{
public:

  class SplitEdge;
  class SplitFace;
  class SplitVertex;

  struct MyUsedTypes : public  UsedTypes<	    Use<SplitVertex>::AsVertexType,
                                           Use<SplitEdge>     ::AsEdgeType,
                                           Use<SplitFace>     ::AsFaceType>{};

  class SplitVertex  : public Vertex<MyUsedTypes, vertex::Coord3f, vertex::Normal3f, vertex::BitFlags  >{};
  class SplitFace    : public Face< MyUsedTypes, face::VertexRef,  face::Normal3f, face::Mark, face::FFAdj, face::BitFlags > {};
  class SplitEdge    : public Edge<MyUsedTypes>{};
  class SplitMesh    : public tri::TriMesh< std::vector<SplitVertex>, std::vector<SplitFace> , std::vector<SplitEdge>  > {};

public:
template <class MeshType>
static void Mesh(MeshType &bm, MeshType &pm, MeshType &out)
{
  SplitMesh om0,om1;

  int insertedNum=0;
  tri::UpdateFlags<MeshType>::VertexClearV(pm);
  tri::UpdateBounding<MeshType>::Box(bm);
  float maxdist = bm.bbox.Diag()/100.0;
  const float snapThr = bm.bbox.Diag()/200.0;
  float founddist;
  Point3f cp;

  SplitMesh *curM = &om1;
  SplitMesh *outM = &om0;
    tri::Append<SplitMesh,MeshType>::Mesh(*curM,bm);
  int step=0;
  while(insertedNum<pm.vn)
  {
    assert(outM != curM);
    GridStaticPtr<SplitFace> myGrd;
    qDebug("\nIteration %i \n",step);
    qDebug("Mesh to refine has %i faces\n",curM->fn);
    tri::UpdateNormals<SplitMesh>::PerFaceNormalized(*curM);
    tri::UpdateBounding<SplitMesh>::Box(*curM);
    tri::UpdateTopology<SplitMesh>::FaceFace(*curM);
    myGrd.Set(curM->face.begin(),curM->face.end(),100);

    tri::FaceTmark<SplitMesh> markerFunctor(curM);
    face::PointDistanceBaseFunctor<float> PDistFunct;

    tri::Allocator<SplitMesh>::AddVertices(*outM,curM->vn);
    for(size_t i=0;i<curM->vn;++i)
      outM->vert[i].P() = curM->vert[i].cP();

    tri::UpdateFlags<SplitMesh>::FaceClearV(*curM);
    // Main loop over all the points to be inserted
    for(size_t i=0;i<pm.vn;++i) if(!pm.vert[i].IsV())
    {
      Point3f qp = pm.vert[i].cP(); // search the face nearest to the the i-th vertex of the point mesh;
      typename SplitMesh::FacePointer fp=myGrd.GetClosest(PDistFunct,markerFunctor,qp,maxdist,founddist,cp);
      assert(fp); // we have built our point set with point onto the surface so it should always find a correct face
      if(!(fp->IsV()))  // we copy and split un-visited faces onto a new mesh;
        {
         int edgeInd;
//         if(SnapToVertex(*fp,qp,snapThr))
//         {
//           pm.vert[i].SetV();
//           pm.vert[i].P()=qp;
//           insertedNum++;
//         }
//         else
//         if(SnapToEdge(*fp,qp,edgeInd,snapThr))
//         {
//           if(!fp->FFp(edgeInd)->IsV())
//           {
//           pm.vert[i].P()=qp;
//           fp->SetV();
//           fp->FFp(edgeInd)->SetV();
//           pm.vert[i].SetV();
//           insertedNum++;
//           FaceType *newfp = &*tri::Allocator<MeshType>::AddFaces(*outM,4);
//           tri::Allocator<MeshType>::AddVertices(*outM,1);

//           int od0 = tri::Index(*curM,fp->V0(edgeInd));
//           int od1 = tri::Index(*curM,fp->V1(edgeInd));
//           int nd1 = tri::Index(*curM,fp->V2(edgeInd));
//           int nd0 = tri::Index(*curM,fp->FFp(edgeInd)->V2(fp->FFi(edgeInd)));
//           outM->vert.back().P()=qp;
//           int nv = outM->vn-1;

//           newfp[0].V(0)=&outM->vert[nd0];
//           newfp[0].V(1)=&outM->vert[od1];
//           newfp[0].V(2)=&outM->vert[nv];

//           newfp[1].V(0)=&outM->vert[od1];
//           newfp[1].V(1)=&outM->vert[nd1];
//           newfp[1].V(2)=&outM->vert[nv];

//           newfp[2].V(0)=&outM->vert[nd1];
//           newfp[2].V(1)=&outM->vert[od0];
//           newfp[2].V(2)=&outM->vert[nv];

//           newfp[3].V(0)=&outM->vert[od0];
//           newfp[3].V(1)=&outM->vert[nd0];
//           newfp[3].V(2)=&outM->vert[nv];

//         }
//         }
//         else
         {
          fp->SetV();
          pm.vert[i].SetV();
          insertedNum++;
          int v0 = tri::Index(*curM,fp->V(0));
          int v1 = tri::Index(*curM,fp->V(1));
          int v2 = tri::Index(*curM,fp->V(2));

          SplitFace *fp = &*tri::Allocator<SplitMesh>::AddFaces(*outM,3);
          tri::Allocator<SplitMesh>::AddVertices(*outM,1);

          outM->vert.back().P()=qp;
          int nv = outM->vn-1;

          fp[0].V(0)=&outM->vert[v0];
          fp[0].V(1)=&outM->vert[v1];
          fp[0].V(2)=&outM->vert[nv];

          fp[1].V(0)=&outM->vert[v1];
          fp[1].V(1)=&outM->vert[v2];
          fp[1].V(2)=&outM->vert[nv];

          fp[2].V(0)=&outM->vert[v2];
          fp[2].V(1)=&outM->vert[v0];
          fp[2].V(2)=&outM->vert[nv];
         }
        }
    }
    printf("Inserted %i vertices\n",insertedNum);
    int insertedFaceNum=0;
    // Second loop: add all the non splitted faces
    for(typename SplitMesh::FaceIterator fi = curM->face.begin() ;fi != curM->face.end();++fi)
      if(!fi->IsV())
      {
        insertedFaceNum++;
        SplitFace *fp = &*tri::Allocator<SplitMesh>::AddFaces(*outM,1);
        for(int jj=0;jj<3;++jj)
          fp->V(jj) = &outM->vert[ tri::Index(*curM,fi->V(jj))];
      }

//    ImproveMesh(*outM);

    curM = outM;
    outM = (curM==&om0)?&om1:&om0 ;
    outM->Clear();
    step++;
  }

  tri::Append<MeshType,SplitMesh>::Mesh(out,*curM);
}

};
}// end namespace vcg
#endif // SPLIT_H
